/*
Package logfunk provides a concise, pluggable API for structured logging.

logfunk uses custom type arguments to set additional logging options.

See README.md for details.
*/
package logfunk

import (
	"io"
	"os"
	"runtime"
	"sort"
	"strconv"
	"strings"

	"git.pixeltamer.net/gopub/timestamp"
	"github.com/davecgh/go-spew/spew"
)

// Level specifies a severity level for a log message. Possible values are Spam, Debug, Info, Warn, Error and Fatal.
type Level int

// MinLevel specifies the minimum severity level for a log message not to be ignored.
type MinLevel int

// KeyValue captures a named value for structured logging
type KeyValue struct {
	Key   string
	Value string
}

// Flags should be set at the root log function and modify the verboseness.
type Flags uint32

// Module is a typed string capturing a module name for logging. Stacked Module names are concatenated with dots in the output.
type Module string

// Extractor values are special log arguments that cause context information to be returned from a log function.
type Extractor int

const (
	// ExtractSink has to be followed by a *LogSink argument that gets set to the parent LogSink instance for this log function.
	ExtractSink = Extractor(0)
	// ExtractDefaults has to be followed by a *[]interface{} argument that gets set to the default parameters for this log function.
	ExtractDefaults = Extractor(1)
)

// The LogSink interface processes LogEntry values and provides default parameters. It is implemented by all logfunk.F returned from logfunk.MakeF()
type LogSink interface {
	LogCommit(le LogEntry)
	LogDefaults() []interface{}
}

// A LogEntry captures all information for a single log line.
type LogEntry struct {
	Time   timestamp.TS
	Level  Level
	Module string
	Text   string
	Flags  Flags
	Attr   map[string]string
}

// Level constants
const (
	Spam Level = iota
	Debug
	Info
	Warn
	Error
	Fatal
)

var (
	logLevel2String   = [6]string{"Spam", "Debug", "Info", "Warn", "Error", "Fatal"}
	logLevel2AlString = [6]string{"Spam ", "Debug", "Info ", "Warn ", "Error", "Fatal"}
)

// Flags
const (
	ShortTime  Flags = 1 << 0
	LongTime   Flags = 0xffffffff ^ ShortTime
	HideLevel  Flags = 1 << 1
	ShowLevel  Flags = 0xffffffff ^ HideLevel
	HideModule Flags = 1 << 2
	ShowModule Flags = 0xffffffff ^ HideModule
	HideCaller Flags = 0xffffffff ^ ShowCaller
	ShowCaller Flags = 1 << 3
)

func (ls Level) String() string {
	return logLevel2String[int(ls)]
}

// AlString returns an aligned string
func (ls Level) AlString() string {
	return logLevel2AlString[int(ls)]
}

// F is a logging function with some twists. All instances of F should be created through MakeF(). F implements the LogSink interface for stacking.
type F func(args ...interface{})

// String implements the Stringer interface
func (le *LogEntry) String() string {
	var l string
	var ls []string
	if le.Flags&ShortTime != 0 {
		ls = append(ls, le.Time.String()[9:19])
	} else {
		ls = append(ls, le.Time.String())
	}
	if le.Flags&HideLevel == 0 {
		ls = append(ls, le.Level.AlString())
	}
	if le.Flags&HideModule == 0 {
		ls = append(ls, le.Module)
	}
	l = strings.Join(ls, " ") + ": " + le.Text
	if le.Attr != nil {
		keys := make([]string, 0)
		for k := range le.Attr {
			keys = append(keys, k)
		}
		sort.Strings(keys)
		l += " ["
		for idx, key := range keys {
			val := le.Attr[key]
			l += key + "="
			l += strconv.Quote(val)
			if idx < len(keys)-1 {
				l += " "
			}
		}
		l += "]"
	}
	return l
}

// dup fmt.Stringer here so we don't have to directly import fmt
type lStringer interface {
	String() string
}

// KV creates a KeyValue object from a key name and an arbitrary value. val is taken verbatim if a string, String()ed if it implements fmt.Stringer and otherwises spew'ed
func KV(key string, val interface{}) KeyValue {
	if s, ok := val.(string); ok {
		return KeyValue{Key: key, Value: s}
	}
	if s, ok := val.(lStringer); ok {
		return KeyValue{Key: key, Value: s.String()}
	}
	return KeyValue{Key: key, Value: spew.Sprint(val)}
}

type iowSink struct {
	w io.Writer
	d []interface{}
}

// LogCommit implements the LogSink interface on the io.Writer wrapper
func (iows *iowSink) LogCommit(le LogEntry) {
	iows.w.Write([]byte(le.String() + "\r\n"))
}

// LogDefaults implements the LogSink interface on the io.Writer wrapper
func (iows *iowSink) LogDefaults() []interface{} {
	return iows.d
}

// WriterSink creates a LogSink wrapper around a writer with optional defaults
func WriterSink(w io.Writer, defaults ...interface{}) LogSink {
	return &iowSink{
		w: w,
		d: append([]interface{}(nil), defaults...),
	}
}

type stdSink struct {
	d []interface{}
}

// LogCommit implements the LogSink interface on the Std wrapper
func (iows *stdSink) LogCommit(le LogEntry) {
	if le.Level >= Warn {
		os.Stderr.Write([]byte(le.String() + "\r\n"))
	} else {
		os.Stdout.Write([]byte(le.String() + "\r\n"))
	}
}

// LogDefaults implements the LogSink interface on the Std wrapper
func (iows *stdSink) LogDefaults() []interface{} {
	return iows.d
}

// StdSink creates a LogSink printing to os.StdOut or os.StdErr depending on Level
func StdSink(defaults ...interface{}) LogSink {
	return &stdSink{
		d: append([]interface{}(nil), defaults...),
	}
}

// GetSink returns the LogSink that was used to create this LogFunc
func (lf F) GetSink() LogSink {
	var ls LogSink
	lf(ExtractSink, &ls)
	return nil
}

// LogCommit submits a LogEntry into the LogFuncs sink
func (lf F) LogCommit(le LogEntry) {
	var ls LogSink
	lf(ExtractSink, &ls)
	if ls != nil {
		ls.LogCommit(le)
	}
}

// LogDefaults retrieves the default values provided to LogMakeFunc originally
func (lf F) LogDefaults() []interface{} {
	var dl []interface{}
	lf(ExtractDefaults, &dl)
	return dl
}

// MakeF creates a new log function on a LogSink (which can be another LogFunc or nil) with optional default values
func MakeF(ls LogSink, defaults ...interface{}) F {
	if ls == nil {
		ls = StdSink()
	}
	appendMod := func(old string, module string) string {
		if old == "" {
			return module
		}
		return old + "." + module
	}
	type ctx struct {
		le   *LogEntry
		minl Level
	}
	apply := func(ctx *ctx, arg interface{}) {
		if v, ok := arg.(Level); ok {
			ctx.le.Level = v
			return
		}
		if v, ok := arg.(MinLevel); ok {
			ctx.minl = Level(v)
			return
		}
		if v, ok := arg.(Module); ok {
			ctx.le.Module = appendMod(ctx.le.Module, string(v))
			return
		}
		if v, ok := arg.(KeyValue); ok {
			if ctx.le.Attr == nil {
				ctx.le.Attr = make(map[string]string)
			}
			ctx.le.Attr[v.Key] = v.Value
		}
		if v, ok := arg.(Flags); ok {
			if v >= 0x80000000 {
				ctx.le.Flags &= v
			} else {
				ctx.le.Flags |= v
			}
		}
	}
	sinkdefaults := ls.LogDefaults()
	alldefaults := []interface{}(nil)
	alldefaults = append(alldefaults, sinkdefaults...)
	alldefaults = append(alldefaults, defaults...)
	return func(args ...interface{}) {
		le := LogEntry{
			Time:   timestamp.Now(),
			Level:  Info,
			Module: "",
			Flags:  0,
		}
		ctx := ctx{le: &le, minl: Spam}
		commit := false
		for i := 0; i < len(alldefaults); i++ {
			apply(&ctx, alldefaults[i])
		}
		for i := 0; i < len(args); i++ {
			arg := args[i]
			if s, ok := arg.(string); ok {
				le.Text = spew.Sprintf(s, args[i+1:]...)
				commit = true
				break
			} else {
				if e, ok := arg.(Extractor); ok {
					i++
					if i < len(args) {
						switch e {
						case ExtractDefaults:
							if dst, ok := args[i].(*[]interface{}); ok {
								*dst = append([]interface{}(nil), alldefaults...)
							}
						case ExtractSink:
							if dst, ok := args[i].(*LogSink); ok {
								*dst = ls
							}
						}
					}
					continue
				}
				apply(&ctx, arg)
			}
		}
		if commit {
			if ctx.minl <= le.Level {
				if le.Flags&ShowCaller != 0 {
					le.AddCallerAttribute()
				}
				if strings.ContainsRune(le.Text, '\n') {
					lines := strings.Split(le.Text, "\n")
					for _, l := range lines {
						le.Text = l
						ls.LogCommit(le)
					}

				} else {
					ls.LogCommit(le)
				}
			}
		}
	}
}

// AddCallerAttribute sets a "Caller" value to the function and the line number in the source file
func (le *LogEntry) AddCallerAttribute() {
	pcs := make([]uintptr, 60)
	depth := runtime.Callers(3, pcs)
	frames := runtime.CallersFrames(pcs[:depth])
	if frames == nil {
		return
	}
	f, _ := frames.Next()
	if le.Attr == nil {
		le.Attr = make(map[string]string)
	}
	le.Attr["Caller"] = f.Function + "," + strconv.Itoa(f.Line)
}
